#  tests for ldap3-2.3-py35h7fea2e8_0 (this is a generated file);
print('===== testing package: ldap3-2.3-py35h7fea2e8_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import ldap3.core
import ldap3.abstract
import ldap3.operation
import ldap3.protocol
import ldap3.protocol.sasl
import ldap3.protocol.schemas
import ldap3.protocol.formatters
import ldap3.strategy
import ldap3.utils
import ldap3.extend
import ldap3.extend.novell
import ldap3.extend.microsoft
import ldap3.extend.standard
#  --- run_test.py (end) ---

print('===== ldap3-2.3-py35h7fea2e8_0 OK =====');
print("import: 'ldap3'")
import ldap3

