# THIS FILE IS AUTO-GENERATED. PLEASE DO NOT MODIFY# version file for ldap3
# generated on 2017-08-02 16:51:51.553923
# on system uname_result(system='Windows', node='ELITE10GC', release='10', version='10.0.15063', machine='AMD64', processor='Intel64 Family 6 Model 58 Stepping 9, GenuineIntel')
# with Python 3.6.1 - ('v3.6.1:69c0db5', 'Mar 21 2017 18:41:36') - MSC v.1900 64 bit (AMD64)
#
__version__ = '2.3'
__author__ = 'Giovanni Cannata'
__email__ = 'cannatag@gmail.com'
__url__ = 'https://github.com/cannatag/ldap3'
__description__ = 'A strictly RFC 4510 conforming LDAP V3 pure Python client library'
__status__ = '5 - Production/Stable'
__license__ = 'LGPL v3'
